@@ -93,13 +93,13 @@ rstan_config <- function(pkgdir = ".") {
   all_files <- list.files(file.path(pkgdir, "src"), full.names = FALSE)
   # reduce to stan model files
   src_files <- all_files[grepl("*[.](cc|h)$", all_files)]
-  src_files <- src_files[grepl(.stan_prefix(start=TRUE), src_files)]
+  src_files <- src_files[grepl(.stan_prefix(start = TRUE), src_files)]
   # make sure 1st line is "don't edit"
   src_line1 <- sapply(file.path(pkgdir, "src", src_files), readLines, n = 1)
   src_files <- src_files[(src_line1 == .rstantools_noedit("foo.h")) |
                          (src_line1 == .rstantools_noedit("foo.cc"))]
   # stan model names corresponding to inactive stan files
-  rm_names <- gsub(.stan_prefix(start=TRUE), "", src_files)
+  rm_names <- gsub(.stan_prefix(start = TRUE), "", src_files)
   rm_names <- unique(gsub("[.](cc|h)$", "", rm_names))
   rm_names <- rm_names[!(rm_names %in% gsub("[.]stan$", "", stan_files))]
   if (length(rm_names) > 0) {
@@ -143,21 +143,19 @@ rstan_config <- function(pkgdir = ".") {
 # return: whether or not file(s) were successfully added/removed
 .setup_Makevars <- function(pkgdir, add = TRUE) {
   noedit_msg <- .rstantools_noedit("foo")
+  mkv_name <- ifelse(.Platform$OS.type == "windows", "Makevars.win", "Makevars")
   if (add) {
-    acc <- sapply(c("Makevars", "Makevars.win"), function(mkv_name) {
-      makevars <- readLines(.system_file(mkv_name))
-      .add_stanfile(makevars, pkgdir, "src", mkv_name,
+    makevars <- readLines(.system_file(mkv_name))
+    acc <- .add_stanfile(makevars, pkgdir, "src", mkv_name,
                     noedit = TRUE, msg = FALSE, warn = TRUE)
-    })
   } else {
-    acc <- sapply(c("Makevars", "Makevars.win"), function(mkv_name) {
-      noedit_msg <- .rstantools_noedit(mkv_name)
-      mkv_name <- file.path(pkgdir, "src", mkv_name)
-      if(file.exists(mkv_name) &&
-         (readLines(mkv_name, n = 1) == noedit_msg)) {
-        file.remove(mkv_name) # Stan file found.  remove it
-      } else FALSE # no stan file found
-    })
+    noedit_msg <- .rstantools_noedit(mkv_name)
+    mkv_name <- file.path(pkgdir, "src", mkv_name)
+    acc <- FALSE
+    if(file.exists(mkv_name) &&
+        (readLines(mkv_name, n = 1) == noedit_msg)) {
+      acc <-  file.remove(mkv_name) # Stan file found.  remove it
+    }
   }
   acc
 }
@@ -174,14 +172,11 @@ rstan_config <- function(pkgdir = ".") {
   ## path to src/stan_files
   ## stan_path <- file.path(pkgdir, "src", "stan_files")
   # create c++ code
-  # .stanfunction compatibility only available after 2.29, so need
-  # to manually wrap function definitions in functions { } before calling stanc
-  if (grepl("\\.stanfunctions$", file_name) &&
-      (utils::packageVersion('rstan') < 2.29)) {
+  if (grepl("\\.stanfunctions$", file_name)) {
     mod <- readLines(file_name)
-    file_name <- paste0(.basename_noext(file_name), "_wrapped.stanfunctions")
-    cat("functions {", mod, "}", sep = "\n",
-        file = file_name)
+    if (!any(grepl("\\bfunctions(\\s*|)\\{", mod))) {
+      writeLines(c("functions {", mod, "}"), sep = "\n", con = file_name)
+    }
   }
   stanc_ret <- rstan::stanc(file_name, allow_undefined = TRUE,
                             obfuscate_model_name = FALSE,
@@ -196,7 +191,7 @@ rstan_config <- function(pkgdir = ".") {
                     !grepl("generated[[:space:]]quantities[[:space:]]*\\{", stanc_ret$model_code)
   if (only_functions) {
     # file_name is a collection of Stan functions rather than a model
-    cppcode <- rstan::expose_stan_functions(stanc_ret, dryRun = TRUE)
+    cppcode <- suppressWarnings(suppressMessages(rstan::expose_stan_functions(stanc_ret, dryRun = TRUE)))
     cpp_lines <- scan(text = cppcode, what = character(),
                       sep = "\n", quiet = TRUE)
     cpp_lines <- cpp_lines[cpp_lines != "#include <exporter.h>"]
@@ -204,18 +199,24 @@ rstan_config <- function(pkgdir = ".") {
     # Stanc3 gives 'auto' return type for standalone functions, which
     #   causes errors with Rcpp::export, so need to replace the auto
     #   return with the plain type from the main definition
-    if(utils::packageVersion('rstan') >= 2.26) {
+    if(utils::packageVersion('rstan') >= "2.26" && utils::packageVersion('StanHeaders') < "2.33") {
       # Extract line numbers of functions to be exported
       decl_lines = grep("// \\[\\[Rcpp::export]]",cpp_lines) + 1
 
       # Replace auto return type with function plain type
       for(i in 1:length(decl_lines)) {
-        cpp_lines[decl_lines[i]] <- .replace_auto(decl_lines[i],cppcode, cpp_lines)
+        next_decl = ifelse(i == length(decl_lines), length(cpp_lines), decl_lines[i] + 1)
+        cpp_lines[decl_lines[i]] <- .replace_auto(decl_lines[i], next_decl, cppcode, cpp_lines)
       }
     }
+    # The default template parameters emitted by stanc3 can error under some clang versions
+    cpp_lines <- gsub(">* = 0>", ">* = nullptr>", cpp_lines, fixed = TRUE)
+    eigen_incl <- ifelse(utils::packageVersion('StanHeaders') >= "2.31",
+                         "#include <stan/math/prim/fun/Eigen.hpp>",
+                         "#include <stan/math/prim/mat/fun/Eigen.hpp>")
     cat("#include <exporter.h>",
-        "#include <stan/math/prim/mat/fun/Eigen.hpp>",
-        "#include <stan/model/standalone_functions_header.hpp>",
+        eigen_incl,
+        "#include <stan/math/prim/meta.hpp>",
         file = file.path(pkgdir, "src",
                          paste(basename(pkgdir), "types.h", sep = "_")),
         sep = "\n")
@@ -228,7 +229,7 @@ rstan_config <- function(pkgdir = ".") {
                       after = class_declaration - 1L)
     # If stan model generated using stanc3, need to make sure that the rstan files are
     #   included after USE_STANC3 has been defined
-    if(utils::packageVersion('rstan') >= 2.26) {
+    if(utils::packageVersion('rstan') >= "2.26") {
       stanc3_declaration <- grep("#define USE_STANC3", cppcode)
       cppcode <- append(cppcode, values = "#include <rstan/rstaninc.hpp>",
                         after = stanc3_declaration + 1)
@@ -349,44 +350,46 @@ rstan_config <- function(pkgdir = ".") {
 }
 
 # Replace auto return type in function exports with the plain type from the main body.
-.replace_auto <- function(decl_line, cppcode, cpp_lines) {
+.replace_auto <- function(decl_line, next_decl, cppcode, cpp_lines) {
   # Extract the name of function
-  decl <- cpp_lines[decl_line]
-  decl <- gsub("auto ","",decl,fixed=T)
-  decl <- sub("\\(.*","",decl,perl=T)
+  fun_name <- paste0(cpp_lines[decl_line:next_decl], collapse = " ")
+  fun_name <- gsub("auto ", "", fun_name, fixed = TRUE)
+  fun_name <- sub("\\(.*", "", fun_name, perl = TRUE)
 
-  # Replace newlines with blank spaces
-  t3 <- gsub("\n"," ",cppcode,fixed=T)
+  # Depending on the version of stanc3, the standalone functions
+  # with a plain return type can either be wrapped in a struct as a functor,
+  # or as a separate forward declaration
+  struct_name <- paste0("struct ", fun_name, "_functor")
 
-  # Identify code segment containing first declaration of function
-  sf1 <- strsplit(t3,decl,fixed=T)[[1]][1]
-  sf2 <- utils::tail(strsplit(sf1,";",fixed=T)[[1]],1)
+  if (grepl(struct_name, cppcode)) {
+    struct_start <- grep(struct_name, cpp_lines)
+    struct_op_start <- grep("operator()", cpp_lines[-(1:struct_start)])[1] + struct_start
 
-  # Get location of type promotion (if present)
-  promote_start <- regexec("stan::promote_args_t<",sf2)[[1]]
+    rtn_type <- paste0(cpp_lines[struct_start:struct_op_start], collapse = " ")
 
-  if(promote_start > 0) {
+    rm_operator <- gsub("operator().*", "", rtn_type)
+    rm_prev <- gsub(".*\\{", "", rm_operator)
+  } else {
+    # Find first declaration of function (will be the forward declaration)
+    first_decl <- grep(paste0(fun_name,"\\("), cpp_lines)[1]
 
-    str_t <- strsplit(sf2,"")[[1]]
-    promote_end <- promote_start + attr(promote_start,'match.length')
+    # The return type will be between the function name and the semicolon terminating
+    # the previous line
+    last_scolon <- grep(";", cpp_lines[1:first_decl])
+    last_scolon <- ifelse(last_scolon[length(last_scolon)] == first_decl,
+                          last_scolon[length(last_scolon) - 1],
+                          last_scolon[length(last_scolon)])
+    rtn_type_full <- paste0(cpp_lines[last_scolon:first_decl], collapse = " ")
+    rm_fun_name <- gsub(paste0(fun_name, ".*"), "", rtn_type_full)
+    rm_prev <- gsub(".*;", "", rm_fun_name)
 
-    count <- 1
-
-    while(count > 0 & promote_end < length(str_t)) {
-      count <- count + ifelse(str_t[promote_end] == "<", 1,
-                             ifelse(str_t[promote_end] == ">", -1,0))
-      promote_end <- promote_end + 1
-    }
-
-    sf2 <- paste0(c(str_t[1:(promote_start-1)], "double",
-                   str_t[promote_end:length(str_t)]),
-                 collapse = "")
   }
 
+  repl_dbl <- gsub("T([0-9])*__", "double", rm_prev)
+
   # Extract return type declaration and replace promoted scalar
   #  type with double
-  rtn_type <- gsub("template <typename(.*?)> ","",sf2)
-
+  rtn_type <- gsub("template <typename(.*?)> ", "", repl_dbl)
   # Update model code with type declarations
-  gsub("auto ", rtn_type, cpp_lines[decl_line],fixed=T)
+  gsub("auto", rtn_type, cpp_lines[decl_line], fixed = TRUE)
 }
